def generate_mappings(config_key, config_value, other_reference_mappings):
    return {
        "DYNAMODB_TABLE_NAME": config_key,
        "DYNAMODB_HASH_KEY": config_value["hash_key"],
        "DYNAMODB_RANGE_KEY": config_value["range_key"],
        "DYNAMODB_ATTRIBUTES": [
            {"name": key, "type": value}
            for key, value in config_value["attributes"].items()
        ],
        "DYNAMODB_GLOBAL_SECONDARY_INDEXES": [
            {
                "name": key,
                "hash_key": value["hash_key"],
                "range_key": value["range_key"],
                "projection_type": value["projection_type"],
            }
            for key, value in config_value.get("global_secondary_indexes", {}).items()
        ],
    }



import os
import re


def replace_placeholders(data, mapping=None, required=True):
    """
    Recursively replace placeholders in data.
    Supports nested keys when the placeholder name includes dots.
    If a placeholder starts with "var.", it is left unchanged.
    """
    if mapping is None:
        mapping = {}

    def get_value(var_name):
        # This function is only called for placeholders that don't start with "var."
        if "." in var_name:
            parts = var_name.split(".")
            value = mapping
            for part in parts:
                if isinstance(value, dict):
                    value = value.get(part)
                else:
                    value = None
                    break
            if value is None and required:
                raise ValueError(f"❌ Error: Missing required variable '{var_name}'.")
            return value
        else:
            value = mapping.get(var_name, os.getenv(var_name))
            if value is None and required:
                raise ValueError(f"❌ Error: Missing required variable '{var_name}'.")
            return value

    if isinstance(data, str):
        # Check if the entire string is an exact placeholder.
        exact_match = re.fullmatch(r"\$\{([^}]+)\}", data)
        if exact_match:
            var_name = exact_match.group(1)
            # If it's a Terraform variable reference, leave it unchanged.
            if var_name.startswith("var.") or var_name.startswith("local."):
                return data
            return get_value(var_name)

        # Otherwise, replace all occurrences.
        def replacer(match):
            var_name = match.group(1)
            if var_name.startswith("var.") or var_name.startswith("local."):
                return match.group(
                    0
                )  # Leave the Terraform variable reference unchanged.
            return str(get_value(var_name))

        return re.sub(r"\$\{([^}]+)\}", replacer, data)
    elif isinstance(data, dict):
        return {
            key: replace_placeholders(value, mapping, required)
            for key, value in data.items()
        }
    elif isinstance(data, list):
        return [replace_placeholders(item, mapping, required) for item in data]
    return data


def to_terraform_hcl(data, indent=2):
    """
    Convert a Python data structure into Terraform HCL format.
    All string values are wrapped in quotes.
    """

    def format_value(value, level=0):
        spacing = " " * (level * indent)
        if isinstance(value, dict):
            lines = ["{"]
            for key, val in value.items():
                lines.append(f"{spacing}  {key} = {format_value(val, level+1)}")
            lines.append(spacing + "}")
            return "\n".join(lines)
        elif isinstance(value, list):
            # For lists of dicts, pretty-print each item.
            if all(isinstance(item, dict) for item in value):
                lines = ["["]
                for item in value:
                    lines.append(f"{spacing}  {format_value(item, level+1)},")
                lines.append(spacing + "]")
                return "\n".join(lines)
            else:
                items = [format_value(item, level) for item in value]
                return "[ " + ", ".join(items) + " ]"
        elif isinstance(value, str):
            if value.startswith("module."):
                return value
            else:
                return f'"{value}"'
        else:
            return str(value)

    return format_value(data)



import json
import os
import subprocess

import click

from .generator import Generator
from .loader import load_env_file, load_yaml

DEFAULT_YAML_FILE = "terrayml.yml"


def generate_terraform(yaml_file, backend_type):
    """Generate Terraform file based on updated YAML configuration."""
    # yaml_file = yaml_file or DEFAULT_YAML_FILE  # Use default YAML if none specified

    try:
        os.makedirs(".tftemp", exist_ok=True)
        Generator.copy_file(file_path=".env.example")

        if not os.path.exists(yaml_file):
            raise FileNotFoundError(
                f"❌ Error: Target Terrayml Config file '{yaml_file}' not found."
            )
        if not os.path.exists("requirements.txt"):
            raise FileNotFoundError(f"❌ Error: requirements.txt file not found.")

        print(f"📄 Using YAML file: {yaml_file}")
        config = load_yaml(yaml_file)
        environment = config["environment"]
        load_env_file(environment)

        updated_config = Generator.assign_variables(config, os.environ)

        Generator.VARIABLE_MAPPING = {
            "AWS_REGION": updated_config["provider"]["region"],
            "PROJECT_NAME": updated_config["project_name"],
            "PROJECT_CODE": updated_config["project_code"],
            "LOWERED_PROJECT_CODE": updated_config["project_code"].lower(),
            "SERVICE_NAME": updated_config["service_name"],
            "AWS_ACCOUNT_ID": os.environ["AWS_ACCOUNT_ID"],
            "AWS_PROFILE": os.environ["AWS_PROFILE"],
            "TERRAYML_PATH": os.getcwd(),
            "ENVIRONMENT": environment,
        }

        if backend_type == "remote":
            Generator.create_file(output_file="main.tf")
        elif backend_type == "local":
            Generator.create_file(
                output_file="backend_main.tf", name_override="main.tf"
            )
        Generator.create_file(output_file="terraform.tfvars")
        Generator.create_file(output_file="variables.tf")

        generator = Generator(updated_config)
        generator.generate()

    except FileNotFoundError as e:
        print(str(e))
        print("❌ Terraform generation failed due to missing file.")
    except ValueError as e:
        print(str(e))
        print(
            "❌ Terraform generation failed due to missing variables. Fix the missing variables and try again."
        )
    except Exception as e:
        print(str(e))
        print("❌ Terraform generation failed due to unknown error.")


@click.group()
def cli():
    """Terrayml - Terraform Generator from YAML"""
    pass


@click.command()
@click.option(
    "--yaml_file",
    default=DEFAULT_YAML_FILE,
    help="Your terrayml.yml format file. Defaults to terrayml.yml",
)
@click.option(
    "--backend_type",
    default="remote",
    help="This indicates that you are generating a remote backend terraform code base. Other option is 'local'.",
)
def generate(yaml_file, backend_type):
    """Generate Terraform files from a YAML file with environment variables."""
    generate_terraform(yaml_file, backend_type)


@click.command()
@click.option(
    "--aws_profile",
    default="default",
    help="Your AWS profile in .aws/credentials.",
)
def local_api(aws_profile):
    """Change directory to .terrayml and run terraform init."""
    terrayml_path = os.path.join(os.getcwd(), ".terrayml")

    click.echo(
        f"Changing directory to {terrayml_path} and running SAM cli command to start local api..."
    )

    if not os.path.isdir(terrayml_path):
        click.echo("Error: .terrayml directory does not exist!")
        return

    click.echo(f"Running nodemon to watch on app folder")
    nodemon_process = subprocess.Popen(
        [
            "nodemon",
            "--watch",
            "app",
            "--exec",
            "terrayml",
            "plan",
            "--lock=false",
            "-e",
            ".py",
        ],
        cwd=os.getcwd(),
    )

    try:
        subprocess.run(
            [
                "sam",
                "local",
                "start-api",
                "--hook-name",
                "terraform",
                "--profile",
                aws_profile,
            ],
            cwd=terrayml_path,
            check=True,
        )
    except subprocess.CalledProcessError as e:
        click.echo(f"SAM Cli failed: {e}")
    finally:
        nodemon_process.terminate()
        click.echo(f"Nodemon Terminated!")


@click.command()
def init():
    """Change directory to .terrayml and run terraform init."""
    terrayml_path = os.path.join(os.getcwd(), ".terrayml")

    if not os.path.isdir(terrayml_path):
        click.echo("Error: .terrayml directory does not exist!")
        return

    click.echo(f"Changing directory to {terrayml_path} and running terraform init...")

    try:
        subprocess.run(["terraform", "init"], cwd=terrayml_path, check=True)
    except subprocess.CalledProcessError as e:
        click.echo(f"Terraform init failed: {e}")


@click.command()
@click.option(
    "--lock",
    default="true",
    help="Acquire state lock or not.",
)
def plan(lock):
    """Change directory to .terrayml and run terraform plan."""
    terrayml_path = os.path.join(os.getcwd(), ".terrayml")

    if not os.path.isdir(terrayml_path):
        click.echo("Error: .terrayml directory does not exist!")
        return

    click.echo(f"Changing directory to {terrayml_path} and running terraform plan...")

    try:
        if lock == "true":
            subprocess.run(["terraform", "plan"], cwd=terrayml_path, check=True)
        elif lock == "false":
            subprocess.run(
                ["terraform", "plan", "-lock=false"], cwd=terrayml_path, check=True
            )
    except subprocess.CalledProcessError as e:
        click.echo(f"Terraform plan failed: {e}")


@click.command()
def apply():
    """Change directory to .terrayml and run terraform apply."""
    terrayml_path = os.path.join(os.getcwd(), ".terrayml")

    if not os.path.isdir(terrayml_path):
        click.echo("Error: .terrayml directory does not exist!")
        return

    click.echo(f"Changing directory to {terrayml_path} and running terraform apply...")

    try:
        subprocess.run(
            ["terraform", "apply", "-auto-approve"], cwd=terrayml_path, check=True
        )
    except subprocess.CalledProcessError as e:
        click.echo(f"Terraform apply failed: {e}")


@click.command()
def destroy():
    """Change directory to .terrayml and run terraform destroy."""
    terrayml_path = os.path.join(os.getcwd(), ".terrayml")

    if not os.path.isdir(terrayml_path):
        click.echo("Error: .terrayml directory does not exist!")
        return

    click.echo(
        f"Changing directory to {terrayml_path} and running terraform destroy..."
    )

    try:
        subprocess.run(
            ["terraform", "destroy", "-auto-approve"], cwd=terrayml_path, check=True
        )
    except subprocess.CalledProcessError as e:
        click.echo(f"Terraform destroy failed: {e}")


cli.add_command(generate)
cli.add_command(init)
cli.add_command(plan)
cli.add_command(apply)
cli.add_command(destroy)
cli.add_command(local_api)

if __name__ == "__main__":
    cli()



def generate_mappings(config_key, config_value, other_reference_mappings):
    return {
        "COGNITO_IDENTITY_POOL_NAME": config_key,
        "COGNITO_IDENTITY_POOL_ALLOW_UNAUTH_IDS": str(
            config_value.get("allow_unauthenticated_identities", True)
        ).upper(),
        "COGNITO_IDENTITY_POOL_IDPS": [
            {
                "user_pool_name": identity_provider["user_pool_name"],
                "client_name": identity_provider["client_name"],
                "server_side_token_check": str(
                    identity_provider.get("server_side_token_check", False)
                ).upper(),
            }
            for identity_provider in config_value.get("cognito_identity_providers", [])
        ],
    }



def generate_mappings(config_key, config_value, other_reference_mappings):
    return {
        "COGNITO_USER_POOL_NAME": config_key,
        "COGNITO_ALLOW_ADMIN_CREATE_USER_ONLY": str(
            config_value.get("allow_admin_create_user_only", False)
        ).upper(),
        "COGNITO_ALIAS_ATTRIBUTE": config_value["alias_attributes"],
        "COGNITO_AUTO_VERIFIED_ATTRIBUTES": config_value["auto_verified_attributes"],
        "COGNITO_DELETION_PROTECTION": str(
            config_value.get("deletion_protection", False)
        ).upper(),
        "COGNITO_DEVICE_CONFIGURATION": {
            "challenge_required_on_new_device": str(
                config_value.get("device_configuration", {}).get(
                    "challenge_required_on_new_device", False
                )
            ).upper(),
            "device_only_remembered_on_user_prompt": str(
                config_value.get("device_configuration", {}).get(
                    "device_only_remembered_on_user_prompt", False
                )
            ).upper(),
        },
        "COGNITO_MFA_CONFIGURATION": config_value.get("mfa_configuration", "OPTIONAL"),
        "COGNITO_SOFTWARE_TOKEN_MFA_ENABLE": str(
            config_value.get("software_token_mfa_configuration_is_enabled", False)
        ).upper(),
        "COGNITO_PASSWORD_POLICY": {
            "minimum_length": config_value.get("password_policy", {}).get(
                "minimum_length", 12
            ),
            "require_lowercase": str(
                config_value.get("password_policy", {}).get("require_lowercase", True)
            ).upper(),
            "require_numbers": str(
                config_value.get("password_policy", {}).get("require_numbers", True)
            ).upper(),
            "require_symbols": str(
                config_value.get("password_policy", {}).get("require_symbols", True)
            ).upper(),
            "require_uppercase": str(
                config_value.get("password_policy", {}).get("require_uppercase", True)
            ).upper(),
            "temporary_password_validity_days": config_value.get(
                "password_policy", {}
            ).get("temporary_password_validity_days", 3),
        },
        "COGNITO_CLIENTS": [
            {
                "name": key,
                "access_token_validity": value.get("access_token_validity", 60),
                "id_token_validity": value.get("id_token_validity", 60),
                "refresh_token_validity": value.get("refresh_token_validity", 30),
            }
            for key, value in config_value.get("clients", {}).items()
        ],
    }



import os
import yaml
from dotenv import load_dotenv

def load_yaml(yaml_file):
    """Load and process the YAML configuration file with environment variables."""
    if not os.path.exists(yaml_file):
        raise FileNotFoundError(f"Error: {yaml_file} not found.")
    
    with open(yaml_file, "r") as file:
        raw_config = yaml.safe_load(file)

    return raw_config

def load_env_file(env_name):
    """Load environment variables from the corresponding .env file and validate required variables."""
    env_file = f"{env_name}.env"
    
    if not os.path.exists(env_file):
        raise FileNotFoundError(f"❌ Error: Environment file '{env_file}' not found.")

    required_vars = get_required_variables()
    
    with open(env_file, "r") as file:
        for line in file:
            if line.strip() and not line.startswith("#"):
                key, value = line.split("=", 1)
                os.environ[key.strip()] = value.strip()

    missing_vars = [var for var in required_vars if os.getenv(var) is None]
    
    if missing_vars:
        raise ValueError(f"❌ Missing required environment variables: {', '.join(missing_vars)}. "
                         f"Ensure they are set in {env_file}.")

    print(f"✅ Loaded environment variables from '{env_file}' successfully.")

def get_required_variables():
    """Read required variables from .env.example file."""
    example_file = ".env.example"
    
    if not os.path.exists(example_file):
        raise FileNotFoundError(f"❌ Error: Missing '.env.example' file, which defines required environment variables.")

    required_vars = []
    
    with open(example_file, "r") as file:
        for line in file:
            if line.strip() and not line.startswith("#"):
                key = line.split("=")[0].strip()
                required_vars.append(key)

    return required_vars



def generate_mappings(config_key, config_value, other_reference_mappings):
    return {
        "LAMBDA_SERVICE_NAME": other_reference_mappings["SERVICE_NAME"],
        "LAMBDA_FUNCTION_NAME": config_key,
        "LAMBDA_FUNCTION_DESCRIPTION": config_value.get("description", ""),
        "LAMBDA_HANDLER": config_value["handler"],
        "LAMBDA_RUNTIME": config_value.get("runtime", "python3.11"),
        "LAMBDA_MEMORY_SIZE": config_value.get("memory_size", 128),
        "LAMBDA_TIMEOUT": config_value.get("timeout", 30),
        "LAMBDA_VPC_SUBNET_IDS": config_value.get("subnet_ids", []),
        "LAMBDA_VPC_SECURITY_GROUP_IDS": config_value.get("security_group_ids", []),
        "LAMBDA_VARIABLES": config_value.get("variables", {}),
    }


import importlib
import json
import os
import shutil

import pkg_resources

from .utilities import replace_placeholders, to_terraform_hcl


class Generator:
    VARIABLE_MAPPING = {}

    def __init__(self, config):
        self.config = config

    @classmethod
    def assign_variables(cls, file, variables, required=True):
        return replace_placeholders(file, variables, required=required)

    @classmethod
    def create_file(cls, output_file, name_override=None):

        template_file = f"terrayml/file_templates/{output_file}.txt"
        template_file = pkg_resources.resource_filename("terrayml", template_file)
        if not os.path.exists(template_file):
            print(f"❌ Error: Terraform template '{template_file}' not found.")
            return

        with open(template_file, "r") as file:
            template = file.read()

        terraform_code = replace_placeholders(
            template, cls.VARIABLE_MAPPING, required=True
        )

        generated_tf_folder_path = ".terrayml"
        os.makedirs(generated_tf_folder_path, exist_ok=True)
        generated_tf_path = os.path.join(generated_tf_folder_path, output_file)
        if name_override:
            generated_tf_path = os.path.join(generated_tf_folder_path, name_override)

        with open(generated_tf_path, "w") as file:
            file.write(terraform_code)

        print(f"✅ Terraform file '{output_file}' generated successfully.")

    @classmethod
    def delete_file(cls, file):
        file_path = f".terrayml/{file}"
        if os.path.exists(file_path):
            os.remove(file_path)
            print(f"✅ Terraform file '{file}' removed successfully.")

    @classmethod
    def create_module(cls, module):
        source_module = f"terrayml/modules/{module}"
        source_module = pkg_resources.resource_filename("terrayml", source_module)
        os.makedirs(".terrayml/modules", exist_ok=True)
        destination_module = f".terrayml/modules/{module}"
        shutil.copytree(
            source_module,
            destination_module,
            dirs_exist_ok=True,
            ignore=lambda src, names: [name for name in names if name == "__init__.py"],
        )
        print(f"✅ Terraform module '{module}' generated successfully.")

    @classmethod
    def delete_module(cls, module):
        os.makedirs("modules", exist_ok=True)
        destination_module = f".terrayml/modules/{module}"

        if os.path.exists(destination_module):
            shutil.rmtree(destination_module)
            print(f"✅ Terraform module '{module}' removed successfully.")

    @classmethod
    def create_object(cls, object_name, mapping={}):
        template_file = f"terrayml/object_templates/{object_name}.json"
        template_file = pkg_resources.resource_filename("terrayml", template_file)
        with open(template_file, "r") as file:
            template = json.load(file)

        return replace_placeholders(template, mapping, required=True)

    @classmethod
    def copy_file(cls, file_path):
        source_path = pkg_resources.resource_filename("terrayml", file_path)
        shutil.copy(source_path, file_path)

    def generate(self):
        for attr_name in dir(self):
            if attr_name.startswith("_generate"):
                method = getattr(self, attr_name)
                if callable(method) and not isinstance(method, classmethod):
                    method()

    def _generate_lambda_module(self):
        config_name = "lambda_functions"
        object_name = config_name[:-1]
        file_template = "lambda_module"
        module = "lambda"

        if self.config.get(config_name, None):
            object_list = []
            apigw_list = []
            ddb_access_list = []

            mapping_module = importlib.import_module(f"mappings.{object_name}")

            for key, value in self.config[config_name].items():
                variable_mapping = {}
                variable_mapping = mapping_module.generate_mappings(
                    key, value, Generator.VARIABLE_MAPPING
                )

                object_list.append(
                    Generator.create_object(object_name, variable_mapping)
                )

                if value.get("path", None):
                    variable_mapping["LAMBDA_PATH"] = value["path"]
                    variable_mapping["LAMBDA_METHOD"] = value["method"]

                    apigw_list.append(
                        Generator.create_object("lambda_apigw", variable_mapping)
                    )
                if value.get("dynamodb_access", False):
                    ddb_access_list.append(
                        Generator.create_object("lambda_access", variable_mapping)
                    )

            Generator.VARIABLE_MAPPING["LAMBDA_FUNCTIONS_LIST"] = to_terraform_hcl(
                object_list
            )
            Generator.VARIABLE_MAPPING["LAMBDA_FUNCTION_WITH_DDB_LIST"] = (
                to_terraform_hcl(ddb_access_list)
            )

            if apigw_list:
                Generator.VARIABLE_MAPPING["ENABLE_APIGW"] = "TRUE"
                Generator.VARIABLE_MAPPING["LAMBDA_FUNCTIONS_WITH_APIGW_LIST"] = (
                    to_terraform_hcl(apigw_list)
                )
                Generator.create_file(output_file="apigw_module.tf")
                Generator.create_module("api_gateway")
            else:
                Generator.VARIABLE_MAPPING["ENABLE_APIGW"] = "FALSE"
                Generator.delete_file("apigw_module.tf")
                Generator.delete_module("api_gateway")

            Generator.create_file(output_file=f"{file_template}.tf")
            Generator.create_module(module)

        else:
            Generator.delete_file(f"{file_template}.tf")
            Generator.delete_module(module)
            Generator.delete_file("apigw_module.tf")
            Generator.delete_module("api_gateway")

    def _generate_dynamodb_module(self):
        config_name = "dynamodb_tables"
        file_template = "dynamodb_module"
        module = "dynamodb"
        object_list_variable = "DYNAMODB_LIST"

        self._simple_module_generation(
            config_name=config_name,
            file_template=file_template,
            module=module,
            object_list_variable=object_list_variable,
        )

    def _generate_s3_module(self):
        config_name = "s3_buckets"
        file_template = "s3_module"
        module = "s3"
        object_list_variable = "S3_BUCKET_LIST"

        self._simple_module_generation(
            config_name=config_name,
            file_template=file_template,
            module=module,
            object_list_variable=object_list_variable,
        )

    def _generate_cognito_module(self):
        config_name = "cognito_user_pools"
        file_template = "cognito_module"
        module = "cognito"
        object_list_variable = "COGNITO_USER_POOL_LIST"

        self._simple_module_generation(
            config_name=config_name,
            file_template=file_template,
            module=module,
            object_list_variable=object_list_variable,
            skip=True,
        )

        config_name = "cognito_identity_pools"
        file_template = "cognito_module"
        module = "cognito"
        object_list_variable = "COGNITO_IDENTITY_POOL_LIST"

        self._simple_module_generation(
            config_name=config_name,
            file_template=file_template,
            module=module,
            object_list_variable=object_list_variable,
        )

    def _simple_module_generation(
        self,
        config_name,
        file_template,
        module,
        object_list_variable,
        skip=False,
    ):
        object_name = config_name[:-1]
        if self.config.get(config_name, None):
            object_list = []

            mapping_module = importlib.import_module(f"mappings.{object_name}")

            for key, value in self.config[config_name].items():
                variable_mapping = {}
                variable_mapping = mapping_module.generate_mappings(
                    key, value, Generator.VARIABLE_MAPPING
                )
                object_list.append(
                    Generator.create_object(object_name, variable_mapping)
                )
            Generator.VARIABLE_MAPPING[object_list_variable] = to_terraform_hcl(
                object_list
            )
            if not skip:
                Generator.create_file(output_file=f"{file_template}.tf")
                Generator.create_module(module)
        else:
            if not skip:
                Generator.delete_file(f"{file_template}.tf")
                Generator.delete_module(module)



from setuptools import find_packages, setup

with open("README.md", "r") as fh:
    description = fh.read()

setup(
    name="terrayml",
    version="0.66",
    author="Warren Ezra Bruce Jaudian",
    author_email="warren.jaudian@ecloudvalley.com",
    packages=find_packages(),
    description="ECV Python Development Package",
    long_description=description,
    long_description_content_type="text/markdown",
    license="MIT",
    py_modules=["terrayml"],
    install_requires=["awscli", "Click", "PyYAML", "python-dotenv"],
    entry_points="""
        [console_scripts]
        terrayml=terrayml.terrayml:cli
    """,
    package_data={
        "terrayml": [
            "terrayml/*",
            ".env.example",
            "py.typed",
        ],
    },
)
